! Module: implot_mod
!
! Authors:  Clive Page - University of Leicester
!          
! Description: Plots EPIC/OM data using PGPLOT and WCS
! 
! Comments: 
! v1.0.1 searches for FITS keywords in main header AND prime image array
! v1.1   fixes bug if no rate/flux info available in source list.
! v1.1.1 changed wcs.f code to get around bug in NAG f90 compiler
! v1.2.2 unchanged, but DISTRIBUTION now "world"
! v1.3   improved pixel range setting
! v1.3.1 fixes bug in DEC labelling, uses pgsbox.f from M.Calabretta
! v1.3.2 improved data scaling, annotation, viewport size.
! v1.3.3 fixed bug in image area causing PGPLOT errors, minor annotation changes.
! v1.3.4 source overlay message often invisible - brought back on screen.
! v1.3.5 improved annotation positioning using pgmtxt not pgptxt
! v1.3.6 fixed crash in heapsort when only 1 source in list
! v1.3.7 fixed msg reporting wrong total no of sources
! v1.3.8 Sort strength by RATE or EP_TOT
! v1.4   Added parameters for pixel range to show: pmin, pmax
! v1.4.1 Cleaned up code; support OM combined source-list overlay
! v1.4.2 Changed annotation so shows "lowest" magnitudes.
! v1.4.3 Fixed -maxsrc bug; added option: numbers, variable size cirles.
! v1.4.4 Fixed bad handing of NaN values in images
! v1.4.5 Oops - special case not handled, fixed this also.
! v1.6   Version with consistent internal/external numbering!
! v1.7   Fixed buf if nfound==0. Improved scaling for nearly constant images
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module implot_mod

  use dal
  use errorHandling
  use metatask_mod
  use param  
  implicit none 
  private 
  public implot
contains 

!---------------------------------------------------------------------
  subroutine implot
!---------------------------------------------------------------------

double precision, parameter  :: pi = 3.141592653589793d0
double precision, parameter  :: dtor = pi/180.0d0
double precision, parameter  :: rtod = 180.0d0/pi

real                         :: blc(2)             ! bottom left corner
double precision             :: cache(4,0:256)
double precision             :: cdelt(2)
integer                      :: ci(7) = -1         ! colours for grid
double precision             :: crpix(2)
double precision             :: crval(2)
character(len=8)             :: ctype(2)
integer                      :: datatype
character(len=80)            :: device
integer                      :: dims(2)
integer                      :: gridtype
double precision             :: dparam(50)=0.0d0 ! parameters for PGWCSL
real                         :: effmin, effmax   ! effective range of image
integer                      :: gcode(2) = 1     ! GCODE arg for PGCRVL
integer                      :: ic
character(len=80)            :: idents(3)     ! annotation for axes, title
integer                      :: ierr          
real, allocatable            :: image(:,:)
type(TaskInfo)               :: info  ! type from metatask_mod
character(len=1024)          :: inset
type(ArrayT)                 :: inArray
type(dataSetT)               :: indataset
integer(kind=int8),  dimension(:,:), pointer :: int8ImageP      
integer(kind=int32), dimension(:,:), pointer :: int32ImageP     
integer(kind=int16), dimension(:,:), pointer :: int16ImageP     
character(len=8)             :: instrument   ! 'OM', 'RGS1' etc.
integer                      :: itf          ! image transfer function
integer                      :: ksl
integer                      :: labctl       ! label control on pgsbox
character(len=80)            :: legend
integer                      :: minx, maxx, miny, maxy
integer                      :: nx, ny
character(len=500)           :: outfile
integer                      :: palette      ! colour palette
integer                      :: pgbeg
real                         :: pmin, pmax   ! user-chosen pixel range
real(kind=single), dimension(:,:), pointer   :: real32ImageP    
real(kind=double), dimension(:,:), pointer   :: real64ImageP    
real, parameter              :: tr(6) = (/ 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 /)
real                         :: trc(2)       ! top right corner
integer                      :: wcs(4)
character(len=10)            :: version = "?"
logical                      :: wcsok
external pgwcsl

! get task version

info = getTaskInfo()
version = versionOf(info)
!
!!print *,'version =', version
! get parameters
!
inset    = stringParameter("set")
device   = stringParameter("device")
outfile  = stringParameter("plotfile")
gridtype = intParameter("grid")
itf      = intParameter("itf")
palette  = intParameter("colour")
pmin     = realParameter("pmin")
pmax     = realParameter("pmax")
!! pmin = -1.0; pmax = -1.0

! open dataset containing image

indataset = dataSet(inset, read)

!
! get handle on image array hdu
!
inArray = array(indataset, "PRIMARY", read)
dims = dimensions(inArray)
nx      = dims(1)
ny      = dims(2)
allocate(image(nx, ny))
!
! get instrument code, and World Coordinate System information if present
! Cannot plot celestial grid or souces positions if WCS absent
!
call getstring("INSTRUME", instrument)
call getdp("CDELT1", cdelt(1))
call getdp("CDELT2", cdelt(2))
call getstring("CTYPE1", ctype(1))
call getstring("CTYPE2", ctype(2))
call getdp("CRPIX1", crpix(1))
call getdp("CRPIX2", crpix(2))
call getdp("CRVAL1", crval(1))
call getdp("CRVAL2", crval(2))
wcsok = (ctype(1)(1:2) == "RA" .and. ctype(2)(1:3) == "DEC")

! Get a pointer to the image, accepting various data types

if (arrayDataType(inArray) == integer8) then
   int8ImageP => int8Array2Data(inArray)
   image = REAL(int8ImageP)
   datatype = 1
elseif (arrayDataType(inArray) == integer16) then
   int16ImageP => int16Array2Data(inArray)
   image = REAL(int16ImageP)
   datatype = 2
elseif (arrayDataType(inArray) == integer32) then
   int32ImageP => int32Array2Data(inArray)
   image = REAL(int32ImageP)
   datatype = 4
elseif (arrayDataType(inArray) == real32) then
   real32ImageP => real32Array2Data(inArray)
   image = real32ImageP
   datatype = -4
elseif (arrayDataType(inArray) == real64) then
   real64ImageP => real64Array2Data(inArray)
   image = real64ImageP
   datatype = -8
end if

if(pmin /= -1.0 .and. pmin /= pmax) then
!
! call eff_range to handle nulls, but ignore results
!
   call eff_range(image, 0.05, 0.01, effmin, effmax)
   effmin = pmin       
   effmax = pmax       ! use the range specified by the user
else 
   if(datatype > 0) then
!
! get effective range of values in image:
! for integer data types, round effmin to next integer value below,
! and ensure effmax is at least 5 and set effmin to zero if 1
!
      call eff_range(image, 0.05, 0.01, effmin, effmax)
      effmin = floor(effmin)
      effmax = max(5.0, effmax)
      if(effmin == 1.0) then
         effmin = 0.0
      end if
   else
      call eff_range(image, 0.05, 0.005, effmin, effmax)
   end if
   if(pmin /= -1.0) effmin = pmin
   if(pmax /= -1.0) effmax = pmax
end if
!
! set up PGPLOT.  Note use PGBEG not PGOPEN because latter
! does not get picked up from library on some installations
!
outfile = trim(outfile) // device
ierr = pgbeg(0, trim(outfile), 1, 1)
if(ierr /= 1) then
  call error("badPGPLOTdevice", "Invalid PGPLOT device: " // trim(outfile))
end if
!!print *,'image size', nx, ny
call im_lims(image, minx, maxx, miny, maxy)
!!print *,'im_lims:', minx, maxx, miny, maxy
call pgsvp(0.1, 0.9, 0.09, 0.93) 
call pgwnad(minx-0.5, maxx+0.5, miny-0.5, maxy+0.5)
call pgsch(0.7)
!
! plot title consiting of input file-name at centre of top
!
ksl = index(inset, "/", .true.)  ! points to last slash
call pgmtxt("T", 1.4, 0.5, 0.5, inset(ksl+1:))
write(legend, "(a,a,a,i2)") "implot v", version
call pgmtxt("T", 1.2, 1.0, 0.0, legend)
call pgscf(2)                 ! better quality Roman font
if(gridtype > 0) then
   call pgbox('BC', 0.0, 0, 'BC', 0.0, 0)
   call pglab(' ', ctype(2), ' ')
else
   call pgbox('BCINT', 0.0, 0, 'BCINT', 0.0, 0)
   call pglab('X', 'Y', " ")
end if
call set_palette(palette)
call pgsitf(itf)
call plot_headers(minx, maxx, miny, maxy)
call pgimag(image, nx, ny, minx, maxx, miny, maxy, effmin, effmax, tr)
!
! plot coordinate grid over the top of the image if requested
!
if(gridtype > 0) then
   if(wcsok) then
      idents(1) = ctype(1)
      idents(2) = " "
      idents(3) = " "
      ic        = -1
      dparam(4)     = crpix(1) 
      dparam(5)     = crpix(2) 
      dparam(6:9)   = (/1.0d0, 0.0d0, 0.0d0, 1.0d0/)
      dparam(10:11) = cdelt
      dparam(20:21) = crval
      dparam(22)    = 999d0
      dparam(23)    = 999d0
      dparam(31)    = 0.0d0
      dparam(32)    = cos(dtor*crval(2)) / sin(dtor*crval(2))
      gcode = gridtype  ! both elements the same
      blc(1) = minx - 0.5
      blc(2) = miny - 0.5
      trc(1) = maxx + 0.5
      trc(2) = maxy + 0.5
      if(gridtype >= 2) ci(1:2) = 4       ! overlay with blue lines
      labctl = 21
      call pgsbox(blc, trc, idents, (/"G","E"/), labctl, 0, ci, gcode, &
        -2.0d0, 0, 0.0d0, 0, 0.0d0, .false., pgwcsl, 16, 4, 50, ctype, &
        wcs, dparam, 256, ic, cache, ierr)
      if(ierr /= 0) then
         call warning("pgcrvl_err", "error in pgsbox")
      end if
   else
     call warning("noWCS", "No WCS keywords") 
   end if
end if
if(booleanParameter("withsrclistset")) then
   if(wcsok) then
      call add_sources(ctype, crval, crpix, cdelt, minx, maxx, miny, maxy,&
                       instrument)
   else
     call warning("noWCS", "No WCS keywords") 
   end if
end if
call pgscf(2)
call pgwedg("LI", 4.9, 2.3, effmin, effmax, " ")
call pgsch(0.5)
call pgmtxt("T", 1.2, -0.07, 1.0, 'cts/pixel')
call pgsch(1.0)
call pgiden
call pgend
deallocate(image)
call release(indataset)

CONTAINS                          ! internal procedures

!-----------------------------------------------------------------------
! getstring: gets FITS string from main dataset or primary array
!-----------------------------------------------------------------------
subroutine getstring(keyword, value)
character(len=*), intent(in)    :: keyword  ! FITS keyword name
character(len=*), intent(out)   :: value    ! value returned or blank
!
if(hasAttribute(indataset, keyword)) then
   value = stringAttribute(indataset, keyword)
else if(hasAttribute(inArray, keyword)) then
   value = stringAttribute(inArray, keyword)
else
   value = " "
end if
end subroutine getstring

!-----------------------------------------------------------------------
! getdp: gets FITS dp from main dataset or primary array
!-----------------------------------------------------------------------
subroutine getdp(keyword, value)
character(len=*), intent(in)  :: keyword  ! FITS keyword name
double precision, intent(out) :: value    ! value returned
!
if(hasAttribute(indataset, keyword)) then
   value = real64Attribute(indataset, keyword)
else if(hasAttribute(inArray, keyword)) then
   value = real64Attribute(inArray, keyword)
else
   value = 0.0d0
end if
end subroutine getdp

!-----------------------------------------------------------------------
! im_lims: computes useful image limits
!-----------------------------------------------------------------------
subroutine im_lims(image, minx, maxx, miny, maxy)
real, intent(in)     :: image(:,:)
integer, intent(out) :: minx, maxx, miny, maxy ! eff image limits
!
integer :: nx, ny, jx, jy 
nx = size(image,1)
ny = size(image,2)
miny = 1
maxy = ny
minx = 1
maxx = nx
do jy = 1, ny
   if(any(image(:,jy) /= 0.0)) EXIT
   miny = jy
end do
do jy = ny, miny, -1
   if(any(image(:,jy) /= 0.0)) EXIT
   maxy = jy
end do
do jx = 1,nx
   if(any(image(jx,:) /= 0.0)) EXIT
   minx = jx
end do
do jx = nx, minx, -1
   if(any(image(jx,:) /= 0.0)) EXIT
   maxx = jx
end do
end subroutine im_lims


!-----------------------------------------------------------------------
! plot_headers: shows FITS headers at top of plot
!-----------------------------------------------------------------------
subroutine plot_headers(minx, maxx, miny, maxy)
integer, intent(in)       :: minx, maxx, miny, maxy ! world coordinate limits
!
integer :: font
character(len=24) :: string, pixels, expos, estring
real :: x, y, dy, csize
!
x = 1.05
y = 0.95
dy = 0.05
call pgqcf(font)
call pgscf(1)
call pgqch(csize)
call pgsch(0.5)
call getstring('INSTRUME', string)
call pgmtxt("RV", x, y, 0.0, string)

call getstring('FILTER', string)
y = y - dy
if(string /= " ") then
   call pgmtxt("RV", x, y, 0.0, trim(string)//' filter')
end if
call getstring('OBJECT', string)
y = y - dy
call pgmtxt("RV", x, y, 0.0, string)
call getstring('OBSERVER', string)
y = y - dy
call pgmtxt("RV", x, y, 0.0, string)
call getstring('DATE-OBS', string)
y = y - dy
call pgmtxt("RV", x, y, 0.0, string(1:16))
call getstring('EXPOSURE', estring)
y = y - dy
if(len_trim(estring) > 0) then
   expos = trim(estring) // ' sec'
   call pgmtxt("RV", x, y, 0.0, expos) 
end if
!
write(pixels, "('pixels (', i0, ':', i0, ',', i0, ':', i0, ')')") &
  minx, maxx, miny, maxy
y = y - 1.5 * dy
call pgmtxt("RV", x, y, 0.0, pixels)
!
call pgscf(font)
call pgsch(csize)
return
end subroutine plot_headers

end subroutine implot

!----------------------------------------------------------------------
! call add_sources
!----------------------------------------------------------------------
subroutine add_sources(ctype, crval, crpix, cdelt, minx, maxx, miny, maxy,&
                       instrument)
character(len=*), intent(in) :: ctype(2)
double precision, intent(in) :: crval(2)
double precision, intent(in) :: crpix(2)
double precision, intent(in) :: cdelt(2)
integer, intent(in) :: minx, maxx, miny, maxy ! image pixel ranges
character(len=*), intent(in) :: instrument
!
integer, pointer             :: band(:)
real                         :: brightest     ! source strength
double precision             :: cel(10)
type(ColumnT)                :: colband
type(ColumnT)                :: coldec
type(ColumnT)                :: colid, colsort
type(ColumnT)                :: colra
integer                      :: colsave
double precision, pointer    :: dec(:)
real, allocatable            :: flux(:)
real                         :: factor        ! circle size factor
integer                      :: fontsave
logical                      :: hasband
double precision             :: imgcrd(2)
integer, pointer             :: id(:)
character(len=12)            :: idents
integer                      :: idmax
integer                      :: ierr          ! error return from wcsfwd
integer                      :: j, k, ksl, n
double precision             :: lin(20)
integer                      :: maxsrc
character(len=50)            :: message
integer                      :: nsources, nfound
integer                      :: noutside      ! count of sources outside image
integer                      :: nshown        ! count of sources shown
integer, allocatable         :: nrec(:)
logical                      :: numbers       ! show numbers by cicles on plot
integer                      :: offset
double precision             :: phi           ! unused
double precision             :: pixcrd(2)     ! pixel coordinates
double precision             :: prj(0:20)     ! internal use in wcsfwd
double precision, pointer    :: ra(:)
real                         :: radiusmax     ! max radius of circles (pixels)
real                         :: radius        ! actual radius to be used
real, pointer                :: rate(:)
character(len=20)            :: ratecol       ! name of column of rates
character(len=8), parameter  :: rate_name(8) = (/ "RATE    ", "EP_TOT  ", &
  "V_INS   ", "B_INS   ", "U_INS   ", "UVW1_INS", "UVW2_INS", "UVM2_INS"/)
character(len=10), parameter :: rec_name(5) = (/ "ML_ID_SRC ", &
  "BOX_ID_SRC", "SRCNUM    ", "SRC_NUM   ", "SRCID     " /)
character(len=1024)          :: sourcefile
type(dataSetT)               :: sourceset
type(TableT)                 :: sourcetable
character(len=12)            :: string, tstring
character(len=200)           :: text
double precision             :: theta         ! unused
integer                      :: thickness, thicksave
logical                      :: varysize      ! cirle size prop to brighness
real                         :: x, y          ! pixel coordinates
integer                      :: wcs(0:3)
double precision             :: world(2)
!
sourcefile  = stringParameter("srclistset")
sourceset   = dataSet(sourcefile, read)
sourcetable = table(sourceset, "SRCLIST")
numbers     = booleanParameter("numbers")
varysize    = booleanParameter("varysize")
colra       = column(sourcetable, "RA", read)
coldec      = column(sourcetable, "DEC", read)
!
! find the name of a suitable count-rate or magnitude column to use when
! finding the top N sources to plot
!
do j = 1, size(rate_name)
   if(hasColumn(sourcetable, rate_name(j))) then
      ratecol   = rate_name(j)
      colsort   = column(sourcetable, ratecol, read)
      EXIT
   end if
end do
if(j > size(rate_name)) then
   call error("no_rate", "no rate/magnitude column")
end if
ra     => real64data(colra)
dec    => real64data(coldec)
rate   => real32data(colsort)
!
! find best column for annotating the source on the plot
!
do j = 1, size(rec_name)
   if(hasColumn(sourcetable, rec_name(j))) then
      idents = rec_name(j)
      colid  = column(sourcetable, idents, read)
      id     => int32data(colid)
      EXIT
   end if
end do
if(j > size(rec_name)) then
!
! no suitable column, report row number.
!
   allocate(id(1:size(ra)))
   idents = 'row'
   do j = 1, size(ra)
      id(j) = j
   end do
end if
!
if(hasColumn(sourcetable, "ID_BAND")) then
   colband = column(sourcetable, "ID_BAND", read)
   band => int32data(colband)
   hasband = .true.
else
   hasband = .false.
end if
!
!
maxsrc      = intParameter("maxsrc")
nsources    = numberOfRows(sourcetable)
wcs = (/ 0, 1, 2, 0/)
cel(1:4)  = (/ crval(1), crval(2), 999.0d0, 999.0d0 /)
cel(5:10) = 0.0d0
lin(1:11) = (/ 0.0d0, 2.0d0, 2.0d0, crpix(1), crpix(2), 1.0d0, 0.0d0, &
             0.0d0, 1.0d0, cdelt(1), cdelt(2) /)
call pgqci(colsave)
call pgsci(7)    ! plot circles colour 7 = yellow
call pgqcf(fontsave)
call pgscf(1)    ! basic font
call pgsfs(2)    ! hollow circles
prj  =  0.0d0
thickness = intParameter("thick")
call pgqlw(thicksave)
call pgslw(thickness)      ! set line thickness
radiusmax = realParameter("radius")
!
! scan the entire table finding source-number, rate, ra, dec
! but if(hasband) then only use ID_BAND==0
!
idmax  = 0
nfound  = 0
allocate(flux(1:nsources), nrec(1:nsources))
!! open(unit=9, file='implot.out', status='replace')
do j = 0, nsources-1
   if(hasband) then
      if(band(j) /= 0) CYCLE
   end if
   if(.not.(isNull(colsort, j) .or. rate(j) >= 1.0e30)) then
      nfound       = nfound + 1
      nrec(nfound) = j          
      flux(nfound) = rate(j)
!!      write(9,'(2i6,es12.3)') nfound, nrec(nfound), flux(nfound)
   end if
end do
!
! now have arrays nrec(1:nfound) and flux(1:nfound)
! sort then into ascending order, moving the nrec values in same way
!
if(nfound > 0) then
   call heapsortr(flux(1:nfound), nrec(1:nfound))
!! write(9,*)'sorted'
!! write(9,'(es12.3,i6)') (flux(j), nrec(j), j=1,nfound)
!
! examine the table, using first maxsrc values for OM data (magnitudes), 
! or the last maxsrc in the table (EPIC, flux)
!
   if(instrument == "OM") then
      offset = 0
      brightest = flux(1)
   else
      offset = max(0, nfound - maxsrc)
      brightest = flux(nfound)
   end if
end if
!! write(9,*)'brightest=', brightest
noutside = 0
nshown   = 0
do k = 1, min(nfound, maxsrc)
   j = nrec(k + offset)
   world(1) = ra(j)
   world(2) = dec(j)
   call wcsfwd(ctype, wcs, world, crval, cel, phi, theta, prj, imgcrd, &
               lin,   pixcrd, ierr)
   x = real(pixcrd(1)) 
   y = real(pixcrd(2)) 
   if(x < minx .or. x > maxx .or. y < miny .or. y > maxy) then
      noutside = noutside + 1
   else
      nshown = nshown + 1
      if(varysize) then
         if(instrument == "OM") then
            factor = 1.0 + 0.5 * (flux(k+offset)-brightest)
         else
            factor = 1.0 + 0.5 * log10(flux(k+offset)/brightest)
         end if
         radius = radiusmax * max(0.35, min(factor, 1.0))
!! write(9,*) k, j, flux(k+offset), factor, radius
      else
         radius = radiusmax
      end if
      call pgcirc(x, y, radius)            ! draw hollow circle
      write(string, *) id(j)
      idmax = max(idmax, id(j))
      if(numbers) then
         n = verify(string, " ")
         tstring = string(n:)
         call pgptxt(x+14.0, y, 0.0, 0.0, tstring)
      end if
   end if
end do 
if(noutside > 0) then
   write(message, '(i6,1x,a)') noutside, 'source positions outside image'
   call warning('sources_outside', message)
end if
!
! annotate plot at bottom with filename and ident info
!
ksl = index(sourcefile, "/", .true.)     ! set ksl to posn of last slash
call pgslw(thicksave)                    ! restore PGPLOT thickness
if(numbers) then
   write(text, "(a, i0, 10a)") &
      "Circled ", nshown, " brightest ", trim(ratecol), " values in ", &
      trim(sourcefile(ksl+1:)), ", numbers from ", trim(idents)
else
   write(text, "(a, i0, 10a)") &
      "Circled ", nshown, " brightest ", trim(ratecol), " values in ", &
      trim(sourcefile(ksl+1:))
end if
!
! restore default colours for text annotation
!
call pgsci(colsave)  
call pgmtxt("B", 4.4, 0.5, 0.5, text)
call pgscf(fontsave)
deallocate(flux, nrec)
call release(sourceset)
end subroutine add_sources


!-----------------------------------------------------------------------
! eff_range  gets useful displayable range of a real image, resets NaNs
!-----------------------------------------------------------------------
subroutine eff_range(image, qlo, qhi, effmin, effmax)
!effmin, effmax, absmin, absmax)
real, intent(inout)       :: image(:,:)  ! real image to examine
real, intent(in)          :: qlo         ! quantile to ignore at low end
                                         ! e.g. 0.05 to ignore lowest 5%
real, intent(in)          :: qhi         ! quantile to ignore at high end
real, intent(out)         :: effmin      ! effective minimum value
real, intent(out)         :: effmax      ! effective maximum value
!
real :: amin, amax, scale
integer, parameter :: nbins = 30000      ! fairly arbitrary histogram size
integer :: hist(0:nbins), npts, ix, iy, j, n, nskip, nsum
!
amin = minval(image)          ! seems to ignore null values
amax = maxval(image) 
!!print *,'amin, amax=', amin, amax
!!print *,'eff_range quantiles ', qlo, qhi
!!print *,'absolute data range ', amin, ' to ', amax
hist = 0                      ! initialise histogram array
npts = 0                      ! will count non-null values
scale = real(nbins) / (amax - amin + 100.0*epsilon(0.0))
do iy = 1,size(image,2)
   do ix = 1,size(image,1)
      if(image(ix,iy) > amin .and. &
         image(ix,iy) <= amax) then         ! this to ignore nulls
            npts = npts + 1
            n = (image(ix,iy) - amin) * scale
            hist(n) = hist(n) + 1
      else                                 ! set null pixels to min level
         image(ix,iy) = amin
      end if
   end do
end do
!
! now find how far up histogram to go to exceed qlo proportion
! of the pixels.
!
nsum  = 0
nskip = nint(qlo * npts)
!!print *,'nskip=', nskip
do j = 0,nbins
   nsum = nsum + hist(j)
!!   write(9,*) j, nsum
   if(nsum > nskip) then    ! reached threshold, set level
      effmin = amin + j / scale
      EXIT
   end if
end do
!
nsum = 0
nskip = nint(qhi * npts)
!!print *,'upper threshold, nskip=', nskip
do j = nbins, 0, -1
   nsum = nsum + hist(j)
!!   write(9,*) j, nsum
   if(nsum > nskip .or. j <= 2) then
      effmax = amin + j / scale
      EXIT
   end if
end do   
!
! fudge here to stop PGPLOT problem if all values same.
!
if(effmax == effmin) effmax = effmin + 100.0*epsilon(0.0)
!!print *,'effective range ', effmin, ' to ', effmax
return
end subroutine eff_range


!-----------------------------------------------------------------------
! set_palette: Sets a color palette for use by PGIMAG
!-----------------------------------------------------------------------
subroutine set_palette(itype)
integer, intent(in) :: itype    !colour palette: -1=inverse gray, 1=grey, 2=rainbow, 3=heat, 
                                ! 4=IRAF, 5=AIPS, 6=TJP
real, parameter :: al(20) = (/0.0, 0.1, 0.1, 0.2, 0.2, 0.3, 0.3, 0.4, 0.4, 0.5, &
                              0.5, 0.6, 0.6, 0.7, 0.7, 0.8, 0.8, 0.9, 0.9, 1.0/)
real, parameter :: ar(20) = (/0.0, 0.0, 0.3, 0.3, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0, &
                              0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0/)
real, parameter :: ag(20) = (/0.0, 0.0, 0.3, 0.3, 0.0, 0.0, 0.0, 0.0, 0.8, 0.8, &
                              0.6, 0.6, 1.0, 1.0, 1.0, 1.0, 0.8, 0.8, 0.0, 0.0/)
real, parameter :: ab(20) = (/0.0, 0.0, 0.3, 0.3, 0.7, 0.7, 0.7, 0.7, 0.9, 0.9, &
                              0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0/)
real, parameter :: grey(2) = (/0.0, 1.0/)
real, parameter :: invg(2) = (/1.0, 0.0/)
real, parameter :: hl(5) = (/0.0, 0.2, 0.4, 0.6, 1.0/)
real, parameter :: hr(5) = (/0.0, 0.5, 1.0, 1.0, 1.0/)
real, parameter :: hg(5) = (/0.0, 0.0, 0.5, 1.0, 1.0/)
real, parameter :: hb(5) = (/0.0, 0.0, 0.0, 0.3, 1.0/)
real, parameter :: rl(9) = (/-0.5, 0.0, 0.17, 0.33, 0.50, 0.67, 0.83, 1.0, 1.7/)
real, parameter :: rr(9) = (/0.0, 0.0, 0.0, 0.0, 0.6, 1.0, 1.0, 1.0, 1.0/)
real, parameter :: rg(9) = (/0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.6, 0.0, 1.0/)
real, parameter :: rb(9) = (/0.0, 0.3, 0.8, 1.0, 0.3, 0.0, 0.0, 0.0, 1.0/)
real, parameter :: tl(4) = (/0.0, 0.5, 0.5, 1.0/)
real, parameter :: tr(4) = (/0.2, 0.6, 0.6, 1.0/)
real, parameter :: tg(4) = (/0.0, 0.0, 0.5, 1.0/)
real, parameter :: tb(4) = (/1.0, 0.0, 0.0, 0.0/)
real, parameter :: wl(10) = (/0.0, 0.5, 0.5, 0.7, 0.7, 0.85, 0.85, 0.95, &
                              0.95, 1.0/)
real, parameter :: wr(10) = (/0.0, 1.0, 0.0, 0.0, 0.3,  0.8,  0.3,  1.0, &
                              1.0, 1.0/)
real, parameter :: wg(10) = (/0.0, 0.5, 0.4, 1.0, 0.0,  0.0,  0.2,  0.7,  &
                              1.0, 1.0/)
real, parameter :: wb(10) = (/0.0, 0.0, 0.0, 0.0, 0.4,  1.0,  0.0,  0.0, &
                              0.95, 1.0/)
! following designed by Clive Page: black + rainbow + white at top end
real, parameter :: cl(7) = (/0.1,  0.25, 0.43, 0.57, 0.71, 0.8,  1.0/)
real, parameter :: cr(7) = (/0.5,  0.0,  0.0,  0.0,  0.9,  1.0,  1.0/)
real, parameter :: cg(7) = (/0.0,  0.0,  0.8,  1.0,  0.9,  0.3,  1.0/)
real, parameter :: cb(7) = (/0.5,  1.0,  0.8,  0.0,  0.0,  0.3,  1.0/)
!                          black  mag   blue  cyan  grn   yell  red   white

select case(itype)
case(-2)                                             ! inverse rainbow
   call pgctab(rl, rr(9:1:-1), rg(9:1:-1), rb(9:1:-1), 9, 1.0, 0.5)
case(-1) 
   call pgctab(grey, invg, invg, invg, 2, 1.0, 0.5)  ! inverse greyscale
case(1)
   call pgctab(grey, grey, grey, grey, 2, 1.0, 0.5)  ! greyscale
case(2)
   call pgctab(rl, rr, rg, rb, 9, 1.0, 0.5)          ! rainbow
case(3)
   call pgctab(hl, hr, hg, hb, 5, 1.0, 0.5)          ! heat
case(4)
   call pgctab(wl, wr, wg, wb, 10, 1.0, 0.5)         ! weird IRAF
case(5)
   call pgctab(al, ar, ag, ab, 20, 1.0, 0.5)         ! AIPS
case(6)
   call pgctab(tl, tr, tg, tb, 4, 1.0, 0.5)          ! Tim Pearson
case(7)
   call pgctab(cl, cr, cg, cb, 8, 1.0, 0.5)         ! Clive Page
   call pgscr(16, 0.0, 0.0, 0.0)                    ! bottom level is black
end select
return
end subroutine set_palette

!----------------------------------------------------------------------
! heapsortr Stable ascending heapsort of real array with links
!----------------------------------------------------------------------

subroutine heapsortr(array, ilink)
real, intent(inout)    :: array(:) ! returned in ascending order
integer, intent(inout) :: ilink(:) ! can keep track of original order
! Author  Clive Page 1991-OCT-1
! basic heapsort algorithm from "Numerical Recipes"
! but equal items will be returned in order of the link values, to
! produce a "stable" sort.
real :: atemp
integer :: npts, k, ir, i, j, tlink
!
npts = size(array)
if(npts <= 1) RETURN
k = npts/2 + 1
ir = npts
do
   if(k  >  1)then
      k = k - 1
      atemp = array(k)
      tlink = ilink(k)
   else
      atemp = array(ir)
      tlink = ilink(ir)
      array(ir) = array(1)
      ilink(ir) = ilink(1)
      ir = ir - 1
      if(ir == 1) then
         ilink(1) = tlink
         array(1) = atemp
         RETURN
      end if
   end if
   i = k
   j = k + k
   do while(j <= ir)
      if(j < ir)then
         if(array(j) < array(j+1) .or.  &
            (array(j) == array(j+1) .and. &
                ilink(j) < ilink(j+1)))      j = j + 1
      end if
      if(atemp < array(j) .or. (atemp == array(j) .and. &
             tlink < ilink(j))) then
         array(i) = array(j)
         ilink(i) = ilink(j)
         i = j
         j = j + j
      else
         j = ir + 1
      end if
   end do
   array(i) = atemp
   ilink(i) = tlink
end do
end subroutine heapsortr

!----------------------------------------------------------------------
! upcase: function converts letters in string to UPPER case 
!----------------------------------------------------------------------
function upcase(string) result(upper)
character(len=*), intent(in) :: string
character(len=len(string)) :: upper
integer :: j
do j = 1,len(string)
  if(string(j:j) >= "a" .and. string(j:j) <= "z") then
       upper(j:j) = achar(iachar(string(j:j)) - 32)
  else
       upper(j:j) = string(j:j)
  end if
end do
end function upcase

end module implot_mod
